import fnmatch
import os
import re

from ..core.exception import TanitError

__all__ = ["glob", "iglob"]


def glob(client, pathname):
    """
    Return a list of paths matching a pathname pattern.

    The pattern may contain simple shell-style wildcards a la
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.
    """
    return list(iglob(client, pathname))


def iglob(client, pathname):
    """
    Return an iterator which yields the paths matching a pathname pattern.

    The pattern may contain simple shell-style wildcards a la
    fnmatch. However, unlike fnmatch, filenames starting with a
    dot are special cases that are not matched by '*' and '?'
    patterns.
    """
    dirname, basename = os.path.split(pathname)
    if not has_magic(pathname):
        if basename:
            status = client.status(pathname, strict=False)
            if status is not None:
                yield pathname
        else:
            # Patterns ending with a slash should match only directories
            status = client.status(pathname, strict=False)
            if status is not None and status["type"] == "DIRECTORY":
                yield pathname
        return
    if not dirname:
        for name in glob1(client, os.curdir, basename):
            yield name
        return
    # `os.path.split()` returns the argument itself as a dirname if it is a
    # drive or UNC path.  Prevent an infinite recursion if a drive or UNC path
    # contains magic characters (i.e. r'\\?\C:').
    if dirname != pathname and has_magic(dirname):
        dirs = iglob(client, dirname)
    else:
        dirs = [dirname]
    if has_magic(basename):
        glob_in_dir = glob1
    else:
        glob_in_dir = glob0
    for dirname in dirs:
        for name in glob_in_dir(client, dirname, basename):
            yield os.path.join(dirname, name)


# These 2 helper functions non-recursively glob inside a literal directory.
# They return a list of basenames. `glob1` accepts a pattern while `glob0`
# takes a literal basename (so it only has to check for its existence).


def glob1(client, dirname, pattern):
    if not dirname:
        if isinstance(pattern, bytes):
            dirname = bytes(os.curdir, "ASCII")
        else:
            dirname = os.curdir
    try:
        names = client.list(dirname)
    except TanitError:
        return []

    if not _ishidden(pattern):
        names = [x for x in names if not _ishidden(x)]

    return fnmatch.filter(names, pattern)


def glob0(client, dirname, basename):
    if basename == "":
        # `os.path.split()` returns an empty basename for paths ending with a
        # directory separator.  'q*x/' should match only directories.
        status = client.status(dirname, strict=False)
        if status is not None and status["type"] == "DIRECTORY":
            return [basename]
    else:
        status = client.status(os.path.join(dirname, basename), strict=False)
        if status is not None:
            return [basename]
    return []


magic_check = re.compile("[*?[]")


def has_magic(s):
    return magic_check.search(s) is not None


def _ishidden(path):
    return path[0] in (".", b"."[0])



import logging as lg
from threading import Thread

from thrift.protocol import TBinaryProtocol
from thrift.server import TServer
from thrift.transport import TSocket
from thrift.transport import TTransport

from ...thrift.worker.filesystem import LocalFilesystem
from .handler import LocalFileSystemHandler

_logger = lg.getLogger(__name__)


class LocalFileSystemService(object):
    def __init__(self, address="0.0.0.0", port=8989):
        self.bind_address = address
        self.bind_port = port

    def _run(self):
        # Create Service handler
        handler = LocalFileSystemHandler()

        server = TServer.TThreadedServer(
            LocalFilesystem.Processor(handler),
            TSocket.TServerSocket(self.bind_address, self.bind_port),
            TTransport.TBufferedTransportFactory(),
            TBinaryProtocol.TBinaryProtocolFactory(),
            daemon=True,
        )

        server.serve()

    def start(self):

        _logger.info("Stating Tanit local file system service.")
        self.daemon = Thread(target=self._run, args=())
        self.daemon.setDaemon(True)
        self.daemon.start()
        _logger.info(
            "Tanit local file system service started, listening  at %s:%s",
            self.bind_address,
            self.bind_port,
        )



import time

import pytest

from tanit.filesystem.filesystem_factory import FileSystemFactory
from tanit.worker.filesystem.service import LocalFileSystemService

from .test_filesystem import BaseFilesystemTest


@pytest.fixture(scope="session")
def filesystem_local_server():
    filesystem_server = LocalFileSystemService(address="0.0.0.0", port=9989)
    filesystem_server.start()


@pytest.fixture(
    scope="class",
    params=[
        {
            "name": "local",
            "type": "local",
            "address": "127.0.0.1",
            "port": "9989",
        },
        {"name": "local", "type": "local"},
    ],
)
def filesystem(filesystem_local_server, request):
    filesystems_factory = FileSystemFactory.getInstance()
    filesystems_factory.register_filesystem(request.param)
    filesystem = filesystems_factory.get_filesystem(request.param["name"])
    yield filesystem


@pytest.fixture(scope="class")
def test_data(filesystem):
    test_dir = "/tmp/integration-test/dataset-%s" % int(time.time() * 1000)
    filesystem.mkdir(test_dir)
    yield test_dir
    filesystem.delete(test_dir, recursive=True)


class TestLocalFilesystem(BaseFilesystemTest):
    pass



import abc
import logging as lg
import os
import os.path as osp
from datetime import datetime
from hashlib import md5
from threading import Lock

import six

from ....common.core.exception import TanitError
from ....common.model.execution_type import ExecutionType
from ....common.utils import glob
from ....common.utils.utils import str2bool
from .execution_state import ExecutionState

_logger = lg.getLogger(__name__)


class TaskExecution(object):
    """TaskExecution represent the execution flow of a task in the master."""

    def __init__(self, tid, etype, params, job):
        self.state = ExecutionState.SUBMITTED

        self.tid = tid
        self.etype = etype
        self.params = params
        self.job = job
        self.attempts = 1
        self.worker = None

    def on_schedule(self):
        if self.state not in [ExecutionState.SUBMITTED, ExecutionState.SCHEDULED]:
            raise IllegalStateTransitionException(
                "Can not transition from state %s to state %s",
                ExecutionState._VALUES_TO_NAMES[self.state],
                ExecutionState._VALUES_TO_NAMES[ExecutionState.SCHEDULED],
            )
        self.state = ExecutionState.SCHEDULED
        self.job.on_task_schedule(self.tid)

    def on_dispatch(self, worker=None):
        if self.state not in [ExecutionState.DISPATCHED, ExecutionState.SCHEDULED]:
            raise IllegalStateTransitionException(
                "Can not transition from state %s to state %s",
                ExecutionState._VALUES_TO_NAMES[self.state],
                ExecutionState._VALUES_TO_NAMES[ExecutionState.DISPATCHED],
            )
        self.state = ExecutionState.DISPATCHED
        self.worker = worker
        self.job.on_task_dispatch(self.tid)

    def on_start(self):
        if self.state not in [
            ExecutionState.RUNNING,
            ExecutionState.DISPATCHED,
            ExecutionState.FAILED,
        ]:  # NOQA
            raise IllegalStateTransitionException(
                "Can not transition from state %s to state %s",
                ExecutionState._VALUES_TO_NAMES[self.state],
                ExecutionState._VALUES_TO_NAMES[ExecutionState.RUNNING],
            )
        self.state = ExecutionState.RUNNING
        self.job.on_task_start(self.tid)

    def on_finish(self):
        if self.state not in [ExecutionState.RUNNING, ExecutionState.FINISHED]:
            raise IllegalStateTransitionException(
                "Can not transition from state %s to state %s",
                ExecutionState._VALUES_TO_NAMES[self.state],
                ExecutionState._VALUES_TO_NAMES[ExecutionState.RUNNING],
            )
        self.state = ExecutionState.FINISHED
        self.job.on_task_finish(self.tid)

    def on_fail(self):
        if self.state not in [ExecutionState.RUNNING, ExecutionState.FAILED]:
            raise IllegalStateTransitionException(
                "Can not transition from state %s to state %s",
                ExecutionState._VALUES_TO_NAMES[self.state],
                ExecutionState._VALUES_TO_NAMES[ExecutionState.RUNNING],
            )
        self.state = ExecutionState.FAILED
        self.job.on_task_fail(self.tid)

    def on_reset(self):
        self.job.on_task_reset(self.tid)

    def on_retry(self):
        self.attempts += 1
        self.job.on_task_reset(self.tid)

    def _reset(self):
        self.state = ExecutionState.SUBMITTED
        self.worker = None


class IllegalStateTransitionException(Exception):
    pass


@six.add_metaclass(abc.ABCMeta)
class JobExecution(object):
    """
    JobExecution represent the execution flow of a job in the master.

    The job state diagram is as follow:
    ----------------------------------------------------------------------
    |   SUBMITTED --> SCHEDULED --> DISPATCHED --> RUNNING --> FINISHED  |
    |                                                |                   |
    |                                                |                   |
    |                                              FAILED                |
    ---------------------------------------------------------------------|
    """

    def __init__(self, params):

        self.state = ExecutionState.SUBMITTED

        self.submission_time = datetime.now()
        self.schedule_time = None
        self.dispatch_time = None
        self.start_time = None
        self.finish_time = None
        self.execution_time_s = -1

        self.tasks = {}

        self.scheduled_tasks = 0
        self.dispatched_tasks = 0
        self.started_tasks = 0
        self.finished_tasks = 0
        self.failed_tasks = 0

        self.jlock = Lock()

        self.initialize(params)

    @abc.abstractmethod
    def initialize(self, params):
        return

    @abc.abstractmethod
    def setup(self):
        return

    def add_task(self, tid, params):
        self.tasks[tid] = TaskExecution(tid, self.etype, params, self)

    def get_task(self, tid):
        if tid in self.tasks:
            return self.tasks[tid]
        else:
            return None

    def get_tasks(self):
        return list(self.tasks.values())

    def get_state(self):
        return self.state

    def on_task_schedule(self, tid):
        with self.jlock:
            self.scheduled_tasks += 1
            # Only the first task transition the state of the job
            if self.state == ExecutionState.SUBMITTED and self.scheduled_tasks == 1:
                _logger.info("Job [ %s ] execution scheduled.", self.jid)
                self.schedule_time = datetime.now()
                self.state = ExecutionState.SCHEDULED

    def on_task_dispatch(self, tid):
        with self.jlock:
            self.dispatched_tasks += 1
            # Only the first task transition the state of the job
            if self.state == ExecutionState.SCHEDULED and self.dispatched_tasks == 1:
                _logger.info("Job [ %s ] execution dispatched.", self.jid)
                self.dispatch_time = datetime.now()
                self.state = ExecutionState.DISPATCHED

    def on_task_start(self, tid):
        with self.jlock:
            self.started_tasks += 1
            # Only the first task transition the state of the job
            if self.state == ExecutionState.DISPATCHED and self.started_tasks == 1:
                _logger.info("Job [ %s ] execution started.", self.jid)
                self.start_time = datetime.now()
                self.state = ExecutionState.RUNNING

    def on_task_finish(self, tid):
        with self.jlock:
            self.finished_tasks += 1
            # All tasks need to finish (successfully)
            # to consider the job finished
            if self.finished_tasks == len(self.tasks):
                _logger.info("Job [ %s ] execution finished.", self.jid)
                self.state = ExecutionState.FINISHED
                self.finish_time = datetime.now()
                self.execution_time_s = int(
                    (self.finish_time - self.start_time).total_seconds()
                )

    def on_task_fail(self, tid):
        with self.jlock:
            self.failed_tasks += 1
            if self.state != ExecutionState.FAILED:
                self.state = ExecutionState.FAILED
                _logger.info("Job [ %s ] execution failed.", self.jid)

    def on_task_reset(self, tid):
        task = self.tasks[tid]
        with self.jlock:
            if task.state == ExecutionState.SCHEDULED:
                self.scheduled_tasks -= 1
            elif task.state == ExecutionState.DISPATCHED:
                self.scheduled_tasks -= 1
                self.dispatched_tasks -= 1
            elif task.state == ExecutionState.RUNNING:
                self.scheduled_tasks -= 1
                self.dispatched_tasks -= 1
                self.started_tasks -= 1
            elif task.state == ExecutionState.FINISHED:
                self.scheduled_tasks -= 1
                self.dispatched_tasks -= 1
                self.started_tasks -= 1
                self.finished_tasks -= 1
                if self.state == ExecutionState.FINISHED:
                    self.state = ExecutionState.RUNNING
            elif task.state == ExecutionState.FAILED:
                self.scheduled_tasks -= 1
                self.dispatched_tasks -= 1
                self.started_tasks -= 1
                if self.state == ExecutionState.FAILED and self.failed_tasks == 1:
                    self.state = ExecutionState.RUNNING
                self.failed_tasks -= 1

            task._reset()


class UploadJobExecution(JobExecution):
    def initialize(self, params):
        raise NotImplementedError

    def setup(self):
        raise NotImplementedError


class MockJobExecution(JobExecution):
    def initialize(self, params):
        try:
            self.num_tasks = int(params["num_tasks"]) if "num_tasks" in params else 2
            self.num_failures = (
                int(params["num_failures"]) if "num_failures" in params else 0
            )
            self.sleep = float(params["sleep"]) if "sleep" in params else 2.0
        except Exception as e:
            raise JobInitializationException("Failed to parse job parameters.", e)

        if self.num_failures > self.num_tasks:
            self.num_failures = self.num_tasks

        self.jid = "mock-job-%s" % (str(datetime.now().strftime("%d%m%Y%H%M%S")))

        self.etype = ExecutionType.MOCK

    def setup(self):
        for i in range(self.num_failures):
            self.add_task(
                "{}-task-{}".format(self.jid, i),
                {"fail": "True", "sleep": str(self.sleep)},
            )

        for i in range(self.num_tasks - self.num_failures):
            self.add_task(
                "{}-task-{}".format(self.jid, i + self.num_failures),
                {"fail": "False", "sleep": str(self.sleep)},
            )


class CopyJobExecution(JobExecution):
    def initialize(self, params):

        if "src" in params:
            self.src = params["src"]
        else:
            raise JobInitializationException(
                "missing required copy job parameter 'src'"
            )

        if "dst" in params:
            self.dst = params["dst"]
        else:
            raise JobInitializationException(
                "missing required copy job parameter 'dst'"
            )

        if "src_path" in params:
            self.src_path = params["src_path"]
        else:
            raise JobInitializationException(
                "missing required copy job parameter 'src_path'"
            )

        if "dest_path" in params:
            self.dest_path = params["dest_path"]
        else:
            raise JobInitializationException(
                "missing required copy job parameter 'dest_path'"
            )

        self.include_pattern = (
            params["include_pattern"] if "include_pattern" in params else "*"
        )
        self.min_size = int(params["min_size"]) if "min_size" in params else 0
        self.overwrite = (
            str2bool(params["overwrite"]) if "overwrite" in params else True
        )
        self.force = str2bool(params["force"]) if "force" in params else True
        self.checksum = str2bool(params["checksum"]) if "checksum" in params else True
        self.files_only = (
            str2bool(params["files_only"]) if "files_only" in params else True
        )
        self.chunk_size = int(params["chunk_size"]) if "chunk_size" in params else 65536
        self.buffer_size = (
            int(params["buffer_size"]) if "buffer_size" in params else 65536
        )

        self.jid = "job-%s-%s" % (
            md5(
                "{}-{}-{}-{}".format(
                    self.src, self.src_path, self.dst, self.dest_path
                ).encode("utf-8")
            ).hexdigest(),
            str(datetime.now().strftime("%d%m%Y%H%M%S")),
        )

        self.etype = ExecutionType.COPY

    def setup(self):

        from ....filesystem.filesystem_factory import FileSystemFactory

        src = FileSystemFactory.getInstance().get_filesystem(self.src)
        dst = FileSystemFactory.getInstance().get_filesystem(self.dst)

        src_path = self.src_path
        dst_path = self.dest_path
        overwrite = self.overwrite

        # Normalise src and dst paths
        src_path = src.resolve_path(src_path)

        if str(dst_path).endswith("/"):
            dst_path = dst.resolve_path(dst_path)
            dst_name = ""
        else:
            r_path = dst.resolve_path(dst_path)
            dst_path = os.path.dirname(r_path)
            dst_name = os.path.basename(r_path)

        # First, resolve the list of src files/directories to be copied
        copies = [copy_file for copy_file in glob.glob(src, src_path)]

        # need to develop a proper pattern based access function
        if len(copies) == 0:
            raise TanitError(
                "Cloud not resolve source path %s : "
                + "either it does not exist or can not access it.",
                src_path,
            )

        tuples = []
        for copy in copies:
            copy_tuple = dict()

            # filename = osp.basename(copy)
            # dst_base_path =  osp.join( dst_path, filename )
            status = dst.status(dst_path, strict=False)
            # statuses = [status for _, status in dst.list(dst_base_path)]
            if status is None:
                # Remote path doesn't exist.
                raise TanitError("Base directory of %r does not exist.", dst_path)
            else:
                # Remote path exists.
                if status["type"] == "FILE":
                    # Remote path exists and is a normal file.
                    if not overwrite:
                        raise TanitError(
                            "Destination path %r already exists.", dst_path
                        )
                    # the file is going to be deleted and the destination
                    # is going to be created with the same name
                    dst_base_path = dst_path
                else:
                    # Remote path exists and is a directory.
                    dst_name = osp.basename(copy) if dst_name == "" else dst_name
                    status = dst.status(osp.join(dst_path, dst_name), strict=False)
                    if status is None:
                        # destination does not exist, great !
                        dst_base_path = osp.join(dst_path, dst_name)
                        pass
                    else:
                        # destination exists
                        dst_base_path = osp.join(dst_path, dst_name)
                        if not overwrite:
                            raise TanitError(
                                "Destination path %r already exists.", dst_base_path
                            )

                copy_tuple = dict({"src_path": copy, "dst_path": dst_base_path})
            tuples.append(copy_tuple)

        # This is a workaround for a Bug when copying files using a pattern
        # it may happen that files can have the same name:
        # ex : /home/user/test/*/*.py may result in duplicate files
        for i in range(0, len(tuples)):
            for x in range(i + 1, len(tuples)):
                if tuples[i]["dst_path"] == tuples[x]["dst_path"]:
                    raise TanitError(
                        "Conflicting files %r and %r : can't copy both files to %r"
                        % (  # NOQA
                            tuples[i]["src_path"],
                            tuples[x]["src_path"],
                            tuples[i]["dst_path"],
                        )
                    )

        for copy_tuple in tuples:
            # Then we figure out which files we need to copy, and where.
            src_paths = list(src.walk(copy_tuple["src_path"]))
            if not src_paths:
                # This is a single file.
                src_fpaths = [copy_tuple["src_path"]]
            else:
                src_fpaths = [
                    osp.join(dpath, fname)
                    for dpath, _, fnames in src_paths
                    for fname in fnames
                ]

            offset = len(copy_tuple["src_path"].rstrip(os.sep)) + len(os.sep)

            i = 1
            for fpath in src_fpaths:
                self.add_task(
                    "{}-task-{}".format(self.jid, i),
                    params={
                        "src": str(self.src),
                        "dst": str(self.dst),
                        "src_path": str(fpath),
                        "dest_path": str(
                            osp.join(
                                copy_tuple["dst_path"],
                                fpath[offset:].replace(os.sep, "/"),
                            ).rstrip(os.sep)
                        ),
                        "overwrite": str(self.overwrite),
                        "force": str(self.force),
                        "checksum": str(self.checksum),
                        "chunk_size": str(self.chunk_size),
                        "buffer_size": str(self.buffer_size),
                    },
                )
                i += 1


class JobInitializationException(Exception):
    pass



#!/usr/bin/env python

import logging as lg
import time
from threading import Thread

from thrift.protocol import TBinaryProtocol
from thrift.server import TServer
from thrift.transport import TSocket
from thrift.transport import TTransport

from ...thrift.master.service import MasterUserService
from ...thrift.master.service import MasterWorkerService
from ..config.config import MasterConfig
from ..core.master import Master
from ..standalone.master import StandaloneMaster
from .handler import UserServiceHandler
from .handler import WorkerServiceHandler

_logger = lg.getLogger(__name__)


class MasterWorkerServer(Thread):
    def __init__(self, master):
        super(MasterWorkerServer, self).__init__()
        self.master = master

    def configure(self, config):
        self.listen_address = config.bind_address
        self.listen_port = config.worker_service_port
        self.n_threads = config.thrift_threads

    def run(self):
        # Create Service handler
        handler = WorkerServiceHandler(self.master)

        server = TServer.TThreadedServer(
            MasterWorkerService.Processor(handler),
            TSocket.TServerSocket(self.listen_address, self.listen_port),
            TTransport.TBufferedTransportFactory(),
            TBinaryProtocol.TBinaryProtocolFactory(),
            daemon=True,
        )

        # Start Tanit server
        server.serve()


class MasterClientServer(Thread):
    def __init__(self, master):
        super(MasterClientServer, self).__init__()
        self.master = master

    def configure(self, config):
        self.listen_address = config.bind_address
        self.listen_port = config.client_service_port
        self.n_threads = config.thrift_threads

    def run(self):
        # Create Service handler
        handler = UserServiceHandler(self.master)

        server = TServer.TThreadedServer(
            MasterUserService.Processor(handler),
            TSocket.TServerSocket(self.listen_address, self.listen_port),
            TTransport.TBufferedTransportFactory(),
            TBinaryProtocol.TBinaryProtocolFactory(),
            daemon=True,
        )

        # Start Tanit server
        server.serve()


class MasterServer(object):
    def __init__(self, config=None, standalone=False):

        self.config = MasterConfig(config)
        self.config.load()

        self.standalone = standalone
        self.master = Master(self.config) if not standalone else StandaloneMaster()

        self.stopped = False

    def stop(self):
        self.stopped = True

    def start(self):
        # Start master services
        self.master.start()

        _logger.info("Stating Tanit master client server.")

        self.mcserver = MasterClientServer(self.master)
        self.mcserver.configure(self.config)
        self.mcserver.setDaemon(True)
        self.mcserver.start()
        _logger.info(
            "Tanit master client server started, listening  at %s:%s",
            self.mcserver.listen_address,
            self.mcserver.listen_port,
        )

        if not self.standalone:
            _logger.info("Stating Tanit master worker server.")
            self.mwserver = MasterWorkerServer(self.master)
            self.mwserver.configure(self.config)
            self.mwserver.setDaemon(True)
            self.mwserver.start()
            _logger.info(
                "Tanit master worker server started, listening  at %s:%s",
                self.mwserver.listen_address,
                self.mwserver.listen_port,
            )

        try:
            while True:
                if not self.mcserver.is_alive():
                    _logger.error(
                        "Unexpected Tanit master client server exit, stopping."
                    )
                    break
                if not self.standalone and not self.mwserver.is_alive():
                    _logger.error(
                        "Unexpected Tanit worker client server exit, stopping."
                    )
                    break
                if self.stopped:
                    _logger.info("Tanit server stopped, exiting.")
                    break
                # wait for 0.5 seconds
                time.sleep(0.5)
        except (KeyboardInterrupt, SystemExit):
            _logger.info("Received KeyboardInterrupt Signal.")
        except Exception as e:
            _logger.exception("Fatal server exception : %s, exiting", e)
        finally:
            _logger.info("Stopping Tanit server.")
            self.master.stop()
            _logger.info("Tanit server stopped.")



import logging as lg
from datetime import datetime
from threading import Thread

_logger = lg.getLogger(__name__)


class WorkerMonitor(Thread):
    # in seconds
    heartbeat_check_interval = 30

    def __init__(self, worker_manager):
        super(WorkerMonitor, self).__init__()
        self.worker_manager = worker_manager
        self.stopped = False
        self.setDaemon(True)

    def stop(self):
        self.stopped = True

    def run(self):
        while not self.stopped:
            for worker in self.worker_manager.list_live_workers():
                if self.heartbeat_check_interval > 0:
                    # only remote workers are monitored
                    if worker.address is not None and worker.port is not None:
                        elapsed_time = datetime.now() - worker.last_hear_beat
                        if elapsed_time.total_seconds() > self.heartbeat_check_interval:
                            _logger.warning(
                                "Missing worker heartbeat from [ %s ] after %s seconds",  # NOQA
                                worker.wid,
                                self.heartbeat_check_interval,
                            )
                            self.worker_manager.decommission_worker(worker.wid)



import contextlib
import os
import time
from threading import Thread

from tanit.master.client.client import ClientType
from tanit.master.client.client import ThriftClientFactory
from tanit.master.config.config import MasterConfig
from tanit.master.server.server import MasterServer
from tanit.worker.server.server import WorkerServer

from ..resources import conf

config_dir = os.path.dirname(os.path.abspath(conf.__file__))


@contextlib.contextmanager
def master_server():
    server = MasterServer(config=config_dir)
    server_daemon = Thread(target=server.start, args=())
    server_daemon.setDaemon(True)
    server_daemon.start()

    # wait for the server to start
    time.sleep(2.0)

    yield

    server.stop()
    server_daemon.join()


@contextlib.contextmanager
def worker_server():
    server = WorkerServer(config=config_dir)
    server_daemon = Thread(target=server.start, args=())
    server_daemon.setDaemon(True)
    server_daemon.start()

    # wait for the server to start
    time.sleep(2.0)

    yield

    server.stop()
    server_daemon.join()


@contextlib.contextmanager
def worker_client():
    config = MasterConfig(path=config_dir)
    factory = ThriftClientFactory(
        config.worker_service_host, config.worker_service_port
    )
    client = factory.create_client(ClientType.WORKER_SERVICE)

    client.start()
    yield client
    client.stop()


def main(argv=None):
    with master_server():
        with worker_server():
            with worker_client() as client:
                print(len(client.list_workers()))


if __name__ == "__main__":
    main()



from functools import wraps
from threading import Thread


def run_async(func):
    @wraps(func)
    def async_func(*args, **kwargs):
        func_hl = Thread(
            target=func,
            args=args,
            kwargs=kwargs,
            name="Thread_{}".format(func.__name__),
        )
        func_hl.start()
        return func_hl

    return async_func


def str2bool(s):
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    else:
        raise BooleanConversionException("can not convert %s to boolean" % s)


class BooleanConversionException(Exception):
    pass
